#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <pwd.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>

#include <sys/types.h>
#include <sys/param.h>

#include "dvbrc.h"
#include "log.h"

#if 0
/*
 * LNB ID 0 TYPE 1  LOF1 9750000 LOF2 10600000 SLOF 11700000 DISEQCNR 0
 *   SAT ID 192 NAME "ASTRA 1B, 1C, 1E, " LNBID 0 FMIN 10700000 FMAX 12700000
 *     TRANSPONDER ID 0014  SATID 0192 TYPE 0 FREQ 11837000 POL H SRATE 27500000 FEC 3/4
 *       CHANNEL ID fd NAME "WDR FERNSEHEN" SATID 192 TPID 14 SID 6dcf TYPE 0 
 */

static struct lnb_s lnb_0 = {
	.id = 0,
	.type = 0,
	.lof1 = 9750000,
	.lof2 = 10600000,
	.slof = 11700000,
};

static struct satellite_s satellite_192 = {
	.id = 0x192,
	.name = "ASTRA 1B, 1C, 1E, ",
	.fmin = 10700000,
	.fmax = 12700000,
	.lnb = &lnb_0,
};

static struct transponder transponder_0014 = {
	.type = DVB_S,
	.tp.s.id = 0x14,
	.tp.s.type = 0,
	.tp.s.freq = 11837000,
	.tp.s.pol = 'H',
	.tp.s.srate = 27500000,
	.tp.s.satellite = &satellite_192,
};

static struct channel channel_102 = {
	.id = 0x102,
	.name = "Das Erste",
	.sid = 0x6dca,
	.type = 0,
	.transponder = &transponder_0014,
};

static struct channel	channel_fd = {
	.next = &channel_102,
	.id = 0xfd,
	.name = "WDR FERNSEHEN",
	.sid = 0x6dcf,
	.type = 0,
	.transponder = &transponder_0014
};

static struct dvbrc mydvbrc = {
	.chans = &channel_fd,
};

#endif

static struct lnb_s lnb_universal = {
	.id = 0,
	.type = 0,
	.lof1 = 9750000,
	.lof2 = 10600000,
	.slof = 11700000,
};

static struct satellite_s dummysat = {
	.id = 1,
	.name = "SomeSatellite",
	.fmin = 10700000,
	.fmax = 12700000,
	.lnb = &lnb_universal,
};

struct dvbrc *	dvbrc = NULL;

/* removes all occurences of char b from string a. */
void delchar(char * a, char b) {
	char * c; char * d;
	c = a; d = a;
	while (*c != '\0') {
		if (*c != b) {
			*d = *c;
			c++; d++;
		} else {
			c++;
		}
	}
	*d = '\0';
}

struct channel * get_channel_by_id(unsigned int cid) {
	struct channel * c;

	if (!dvbrc)
		return NULL;

	c = dvbrc->chans;
	while (c) {
		if (c->id == cid) {
			return c;
		}
		c = c->next;
	}

	return NULL;
}

int splitintoparts(char * what, char ** whereto, int maxparts) {
	int curpart = 0; int pis = 0;
	while ((*what != '\0') && (*what != '\r') && (*what != '\n')) {
		if (*what == ':') {
			whereto[curpart][pis] = 0;
			pis = 0; curpart++;
			if (curpart >= maxparts) {
				return maxparts+1;
			}
		} else {
			whereto[curpart][pis] = *what;
			pis++;
		}
		what++;
	}
	whereto[curpart][pis] = 0;
	if (pis > 0) {
		return (curpart+1);
	}
	return curpart;
}

#define MAXPARTS 20
#define TMPSSIZE 1024

static int trytoreadchannellist(FILE * f) {
	char rbuf[TMPSSIZE];
	int sparts;
	struct channel * c = NULL;
	struct transponder * t = NULL;
	unsigned int lasttrpid = 1;
	unsigned int chansread = 0;
	unsigned int tmpsid; /* */
	char * splitbuf[MAXPARTS];
	
	if (!dvbrc) {
		fprintf(stderr, "%s:%d: dvbrc data structure not allocated?! This is a BUG\n", __FUNCTION__, __LINE__);
		return 1;
	}
	for (sparts = 0; sparts < MAXPARTS; sparts++) {
		splitbuf[sparts] = (char *)malloc(TMPSSIZE);
	}
	while (!feof(f)) {
		rbuf[0] = 0;
		fgets(rbuf, TMPSSIZE, f);
		delchar(rbuf, '\r'); /* Remove linefeeds, just to make it look */
		delchar(rbuf, '\n'); /* nicer in warnings or debug msgs */
		if (rbuf[0] == ' ') continue; /* our lines never start with a space */
		if (rbuf[0] == '#') continue; /* ignore comments */
		if (rbuf[0] == 0) continue; /* Shortcut for empty lines */
		sparts = splitintoparts(rbuf, splitbuf, MAXPARTS);
		if        (sparts == 13) { /* DVB-T */
			unsigned long tmpfreq; /* frequency */
			unsigned int tmpbw; /* bandwidth */
			unsigned int tmpmodul; /* modulation */
			unsigned int tmptrans; /* transmission */
			unsigned int tmpguard; /* guard interval */
			signed int tmphier; /* hierarchy */
			/* Example:
			 * RTL Television:834000000:INVERSION_AUTO:BANDWIDTH_8_MHZ\
			 *   :FEC_2_3:FEC_1_2:QAM_16:TRANSMISSION_MODE_8K\
			 *   :GUARD_INTERVAL_1_4:HIERARCHY_NONE:337:338:16405
			 * (0)ChannelName:(1)Frequency:(2)Inversion:(3)Bandwidth\
			 *   :(4)FEC1:(5)FEC2:(6)QAM:(7)TransmissionMode\
			 *   :(8)GuardInterval:(9)Hierarchy:(10)vpid:(11)apid:(12)sid
			 */
			/* Frequency */
			tmpfreq = strtoul(splitbuf[1], NULL, 10);
			if (tmpfreq == 0) {
				putlog(0, "Warning: Invalid frequency: %s in line %s (ignored)", splitbuf[1], rbuf);
				continue;
			}
			/* BANDWITH_X */
			if (strncmp(splitbuf[3], "BANDWIDTH_", 10)) {
				putlog(0, "Warning: no BANDWIDTH_ in line in channelconfig (ignored): %s", rbuf);
				continue;
			}
			tmpbw = strtoul(&splitbuf[3][10], NULL, 10);
			/* QAM_X (modulation) */
			if (strncmp(splitbuf[6], "QAM_", 4)) {
				putlog(0, "Warning: no QAM_ in line in channelconfig (ignored): %s", rbuf);
				continue;
			}
			tmpmodul = strtoul(&splitbuf[6][4], NULL, 10);
			/* TRANSMISSION_MODE_XK */
			if (strncmp(splitbuf[7], "TRANSMISSION_MODE_", 18)) {
				putlog(0, "Warning: no TRANSMISSION_MODE_ in line in channelconfig (ignored): %s", rbuf);
				continue;
			}
			tmptrans = strtoul(&splitbuf[7][18], NULL, 10);
			/* GUARD_INTERVAL_1_Y */
			if (strncmp(splitbuf[8], "GUARD_INTERVAL_1_", 17)) {
				putlog(0, "Warning: no GUARD_INTERVAL_1_ in line in channelconfig (ignored): %s", rbuf);
				continue;
			}
			tmpguard = strtoul(&splitbuf[8][17], NULL, 10);
			/* HIERARCHY_X */
			if (strncmp(splitbuf[9], "HIERARCHY_", 10)) {
				putlog(0, "Warning: no HIERARCHY__ in line in channelconfig (ignored): %s", rbuf);
				continue;
			}
			if (!strcmp(&splitbuf[9][10], "NONE")) {
				tmphier = -1;
			} else {
				tmphier = strtoul(&splitbuf[9][10], NULL, 10);
			}
			/* SID */
			tmpsid = strtoul(splitbuf[12], NULL, 10);
			if (tmpsid == 0) {
				putlog(0, "Warning: SID of 0 in line in channelconfig (ignored): %s", rbuf);
				continue;
			}
			
			/* If we reach this we have a full channel definition, and
			 * possibly implicitly a transponder definition. */
			putlog(9, "Read channel/transponder definition: freq=%lu"
				" name=%s sid=%u bw=%u modul=%u trans=%u guard=%u hier=%d",
				tmpfreq, splitbuf[0], tmpsid, tmpbw, tmpmodul, tmptrans, tmpguard, tmphier);
			
			if (!t) {
				t = calloc(1, sizeof(struct transponder));
			} else {
				if (t->tp.t.freq != tmpfreq) {
					t = calloc(1, sizeof(struct transponder));
				}
			}
			if (t->tp.t.id == 0) {
				t->type = DVB_T;
				t->tp.t.id = lasttrpid++;
				t->tp.t.freq = tmpfreq;
				t->tp.t.bandw = tmpbw;
				t->tp.t.modul = tmpmodul;
				t->tp.t.trans = tmptrans;
				t->tp.t.guard = tmpguard;
				t->tp.t.hier = tmphier;
			}
			/* NO CONTINUE! */
		} else if (sparts ==  8) { /* DVB-S */
			unsigned long tmpfreq; /* frequency */
			char tmppol; /* Polarization (uppercase! H|V) */
			unsigned long tmpsrate; /* Symbolrate */
			/* Example:
			 * VOX:12187:h:0:27500:167:136:12060
			 * (0)Name:(1)Frequency:(2)Pol:(3)??:(4)Symbolrate:(5)vpid:(6)apid:(7)sid
			 */
			/* Frequency */
			tmpfreq = strtoul(splitbuf[1], NULL, 10);
			if (tmpfreq == 0) {
				putlog(0, "Warning: Invalid frequency: '%s' in line '%s' (ignored)", splitbuf[1], rbuf);
				continue;
			}
			if (tmpfreq < 900000) { tmpfreq *= 1000; } /* MHz -> KHz */
			/* Polarization */
			if (strlen(splitbuf[2]) != 1) {
				putlog(0, "Warning: Invalid polarization: '%s' in line '%s' (ignored)", splitbuf[2], rbuf);
				continue;
			}
			tmppol = toupper(splitbuf[2][0]);
			if ((tmppol != 'H') && (tmppol != 'V')) {
				putlog(0, "Warning: Invalid polarization: '%s' in line '%s' (ignored)", splitbuf[2], rbuf);
				continue;
			}
			/* Symbolrate */
			tmpsrate = strtoul(splitbuf[4], NULL, 10);
			if (tmpsrate == 0) {
				putlog(0, "Warning: Invalid Symbolrate: '%s' in line '%s' (ignored)", splitbuf[4], rbuf);
				continue;
			}
			if (tmpsrate < 900000) { tmpsrate *= 1000; } /* MHz -> KHz */
			/* SID */
			tmpsid = strtoul(splitbuf[7], NULL, 10);
			if (tmpsid == 0) {
				putlog(0, "Warning: SID of 0 in line in channelconfig (ignored): %s", rbuf);
				continue;
			}
			
			/* If we reach this we have a full channel definition, and
			 * possibly implicitly a transponder definition. */
			putlog(9, "Read channel/transponder definition: freq=%lu"
				" name=%s sid=%u pol=%c srate=%lu",
				tmpfreq, splitbuf[0], tmpsid, tmppol, tmpsrate);
			
			if (!t) {
				t = calloc(1, sizeof(struct transponder));
			} else {
				if ((t->tp.s.freq != tmpfreq) || (t->tp.s.pol != tmppol)) {
					t = calloc(1, sizeof(struct transponder));
				}
			}
			if (t->tp.s.id == 0) {
				t->type = DVB_S;
				t->tp.s.id = lasttrpid++;
				t->tp.s.freq = tmpfreq;
				t->tp.s.satellite = &dummysat;
				t->tp.s.pol = tmppol;
				t->tp.s.srate = tmpsrate;
			}
			/* NO CONTINUE! */
		} else if ((sparts ==  4) && (strcmp(splitbuf[0], "LNB") == 0)) {
			/* LNB definition */
			dummysat.lnb = calloc(1, sizeof(struct lnb_s));
			dummysat.lnb->lof1 = strtoul(splitbuf[1], NULL, 0);
			dummysat.lnb->lof2 = strtoul(splitbuf[2], NULL, 0);
			dummysat.lnb->slof = strtoul(splitbuf[3], NULL, 0);
			putlog(9, "New LNB type learnt: lof1=%lu lof2=%lu slof=%lu",
			       dummysat.lnb->lof1, dummysat.lnb->lof2, dummysat.lnb->slof);
			continue;
		} else {
			putlog(0, "Warning: Unparsable line with %d parameters in channelconfig ignored: %s", sparts, rbuf);
			continue;
		}
		
		/* Now add the channel (independent from transponder type) */
		if (!c) {
			c = calloc(1, sizeof(struct channel));
			dvbrc->chans = c;
		} else {
			c->next = calloc(1, sizeof(struct channel));
			c = c->next;
		}
		c->sid = tmpsid;
		c->id = tmpsid;
		c->name = malloc(strlen(splitbuf[0])+1);
		strcpy(c->name, splitbuf[0]);
		c->transponder = t;
		chansread++;
	}
	
	if (chansread > 0) {
		putlog(1, "Read %u transponders with %u channels from dvbrc",
		          lasttrpid-1, chansread);
	}
	for (sparts = 0; sparts < MAXPARTS; sparts++) {
		free(splitbuf[sparts]);
	}
	return 0;
}

int read_dvbrc(char * fn) {
	char fname[PATH_MAX+1];
	FILE * ccf; /* Channel Config File */

	if (dvbrc) {
		/* Channel data is already defined - do not overwrite it */
		putlog(0, "Warning: Using compiled-in channel definitions instead of .dvbrc!\n");
		return 0;
	}

	if (fn == NULL) {
		struct passwd *p;
		p = getpwuid(geteuid());
		snprintf(fname, PATH_MAX, "%s/.dvbrc", (p->pw_dir) ? p->pw_dir : "");

		ccf = fopen(fname, "r");
		if (ccf == (FILE *)0) {
			putlog(0, "Warning: No %s, retrying in current directory...", fname);
			ccf=fopen(".dvbrc", "r");
			if (ccf == (FILE *)0) {
				perror("Opening .dvbrc");
				return 1;
			}
		}
	} else {
		ccf = fopen(fn, "r");
		if (ccf == (FILE *)0) {
			perror("Opening channel list");
			putlog(0, "Could not open %s - sorry.", fn);
			return 1;
		}
	}

	dvbrc = calloc(1, sizeof(struct dvbrc));

	putlog(3, "Trying to read channel config from %s...", (fn) ? fn : fname);
	if (trytoreadchannellist(ccf)) {
		putlog(0, "Could not read channel list.");
	}

	fclose(ccf);

	return 0;
}
