
#include <sys/poll.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <getopt.h>

#include <event.h>

#include "log.h"

/*
 * Defines
 */

#define MAX_MCAST_PAYLOAD	(1500-40)
#define TS_PACKET_SIZE		188

#define MAXSENDOVERFLOWS	500	/* After this many overflows the connection gets dropped */

#define HTTP_OK_REPLY "HTTP/1.0 200 OK\r\n"\
                      "Content-type: application/octet-stream\r\n"\
                      "\r\n"
#define HTTP_REQUEST_PART1 "GET "
#define HTTP_REQUEST_PART2 " HTTP/1.0\r\n"\
                           "\r\n"

/* Stream Types */
#define ST_HTTP		0
#define ST_MCAST	1
#define ST_FILE		2

#ifndef FILE_MAX
#define FILE_MAX 4096
#endif

/*
 * Types
 */
 
struct unicast_client {
	struct unicast_client *	next;
	int			sockfd;
	struct event		ev;
	unsigned long		sendoverflows;
	char			hadrequest;
};

struct streamresource {
	int			isopen;	/* boolean value - source open or not */
	int			fd;
	struct event		ev;
	int			type;
	char			fname[FILE_MAX]; /* Filename for ST_FILE, or textual IP/host else */
	uint32_t		ipv4;	/* IP for ST_HTTP or ST_MCAST */
	uint16_t		port;	/* Port for ST_HTTP or ST_MCAST */
	char			httpdone; /* Gets set to 1 when we're through with the HTTP crap */
	int			nuccs;	/* Number of UniCast Clients for ST_HTTP */
	struct unicast_client *	ucclients; /* Clients for ST_HTTP */
};

struct stream {
	struct stream *		next;
	struct streamresource	src;
	struct streamresource	dst;
	int			pbufvalid;
	uint8_t 		pbuf[MAX_MCAST_PAYLOAD];
};

/*
 * Global Variables
 */

unsigned int	sndbuf_size = 65536 * 4; /* 256 KB */
int		readpersist = 0; /* 0 if we close the source when there are no listeners. */
struct stream *	streams = NULL;
unsigned int	maxuccs = 5; /* Maximum number of unicast clients */
int		mcastttl = 1;

/*
 * Functions
 */

void openstreamresource(struct streamresource * sr, int issrc, struct stream * parent);

void sigpipehandler(int bla) { /* Dummyhandler for catching the event */
	return; /* It does nothing, but without it our program would die
		 * on receiving a SIGPIPE */
}

void init_sigpipehandler(void) {
	struct sigaction sia;
	sia.sa_handler = sigpipehandler;
	sigemptyset(&sia.sa_mask); /* If we don't do this, we're likely */
	sia.sa_flags = 0;          /* to die from 'broken pipe'! */
	sigaction(SIGPIPE, &sia, NULL);
}

void drop_tcp_connection(struct stream * str, int fd) {
	struct unicast_client * uc = str->dst.ucclients;
	struct unicast_client * prev = NULL;
	
	while (uc) {
		if (uc->sockfd == fd) { /* That's the one. */
			event_del(&uc->ev);
			close(uc->sockfd);
			if (prev == NULL) {
				str->dst.ucclients = uc->next;
			} else {
				prev->next = uc->next;
			}
			free(uc);
			str->dst.nuccs--;
			if ((str->dst.nuccs == 0) && (readpersist == 0)) {
				putlog(1, "Closing upstream fd %d, no more clients", str->src.fd);
				event_del(&str->src.ev);
				close(str->src.fd);
				str->src.isopen = 0;
				str->pbufvalid = 0;
			}
			return;
		}
		prev = uc;
		uc = uc->next;
	}
}

void setfdnonblocking(int fd) {
	int flags;
	flags = fcntl(fd, F_GETFL, 0);
	flags |= O_NONBLOCK|O_NDELAY;
	if (fcntl(fd, F_SETFL, flags) < 0) {
		perror("nonblocking fcntl");
		exit(1);
	}
}

void showhelptxt(char * progname) {
	printf("Syntax: %s [-d level] [-h] (pipe|streamaddr)1 (unicast|multicast)addr1 ... \n", progname);
	printf("-d level       Sets the log level. Higher level gives more output.\n");
	printf("               defaults to 1.\n");
	printf("-b sndbuf      How much bufferspace to request from the OS (SO_SNDBUF)\n");
	printf("-l limit       Limit for number of connections PER STREAM (!).\n");
	printf("               Used for unicast only (there are no connections on mcast)\n");
}

void parseurl(char * what, struct streamresource * sr) {
	char * spl;
	struct hostent *hp;
	uint8_t tmpstr[FILE_MAX];
	
	memset(sr, 0, sizeof(struct streamresource));
	sr->type = ST_HTTP;
	if        (strncmp(what, "file://", 7) == 0) {
		sr->type = ST_FILE;
		what += 7;
	} else if (strncmp(what, "udp://", 6) == 0) {
		sr->type = ST_MCAST;
		what += 6;
	} else if (strncmp(what, "http://", 7) == 0) {
		sr->type = ST_HTTP;
		what += 7;
	}
	strncpy(&sr->fname[0], what, FILE_MAX-1); /* Terminating 0 from memset above */
	strcpy(tmpstr, &sr->fname[0]);
	if (sr->type == ST_FILE) {
		/* We're finished already. */
		return;
	}
	/* For mcast or http: split into hostname/ip and port, and resolve */
	spl = strtok(tmpstr, ":/");
	if (spl == NULL) {
		fprintf(stderr, "Failed to parse URL '%s'", what);
		exit(1);
	}
	hp = gethostbyname(spl);
	if (hp == NULL) {
		fprintf(stderr, "Failed to resolve Host '%s'\n", spl);
		exit(1);
	}
	if ((hp->h_length != sizeof(uint32_t))
	 || (hp->h_addrtype != AF_INET)) {
		fprintf(stderr, "Host '%s' did not resolve to IPv4 addr\n", spl);
		exit(1);
	}
	memcpy((char *)&sr->ipv4, hp->h_addr, hp->h_length);
	spl = strtok(NULL, ":/");
	if (spl == NULL) {
		/* Default Port: 1234. */
		sr->port = 1234;
	} else {
		sr->port = strtoul(spl, NULL, 0);
		while (spl != NULL) {
			/* Read away all tokens so the original string gets restored */
			spl = strtok(NULL, ":/");
		}
	}
}

static inline void send_packet(struct stream * str, uint8_t * buf) {
	int len;
	if        (str->dst.type == ST_MCAST) {
		len = send(str->dst.fd, str->pbuf, str->pbufvalid, MSG_DONTWAIT);
		if (len != str->pbufvalid) {
			putlog(0, "send didn't send all... %d/%d", len, str->pbufvalid);
		}
	} else if (str->dst.type == ST_HTTP) {
		/* Send to all registered clients */
		struct unicast_client * uc = str->dst.ucclients;
		while (uc) {
			if (!uc->hadrequest) {
				uc = uc->next;
				continue;
			}
			len = write(uc->sockfd, str->pbuf, str->pbufvalid);
			if (len != str->pbufvalid) {
				putlog(2, "[fd %05d]: failed send for the %luth time... %d/%d",
					uc->sockfd, uc->sendoverflows, len, str->pbufvalid);
				uc->sendoverflows++;
				if (uc->sendoverflows > MAXSENDOVERFLOWS) {
					struct unicast_client * olduc = uc;
					uc = uc->next;
					putlog(1, "[fd %05d]: Too many overflows, dropping connection.",
						  olduc->sockfd);
					drop_tcp_connection(str, olduc->sockfd);
					continue;
				}
			}
			uc = uc->next;
		}
	} else if (str->dst.type == ST_FILE) {
		len = write(str->dst.fd, str->pbuf, str->pbufvalid);
		if (len != str->pbufvalid) {
			putlog(0, "write didn't write all... %d/%d", len, str->pbufvalid);
		}
	} else {
		/* WTF? */
	}
}

void data_incoming(int fd, short event, void * up) {
	int len;
	struct stream * str;
	int readsize = TS_PACKET_SIZE * 3;
	
	str = (struct stream *)up;
	if (str->src.fd != fd) {
		putlog(0, "!!! %s called with nonmatching data !!!", __FUNCTION__);
		return;
	}
	if (str->src.type == ST_MCAST) {
		readsize = MAX_MCAST_PAYLOAD;
	}
	len = read(fd, &str->pbuf[str->pbufvalid], readsize);
	if (len >= 0) {
		if ((str->src.type == ST_HTTP) && (!str->src.httpdone)) {
			int j;
			for (j = 0; j < (len-4); j++) {
				if (strncmp(&str->pbuf[j+0], "\r\n\r\n", 4) == 0) {
					putlog(1, "HTTP done on upstream fd %d", str->src.fd);
					str->src.httpdone = 1;
					memmove(&str->pbuf[0], &str->pbuf[j+4], (len - (j + 4)));
					len -= (j + 4);
					break;
				}
			}
			if (!str->src.httpdone) {
				for (j = 0; j < (len-2); j++) {
					if (strncmp(&str->pbuf[j+0], "\n\n", 2) == 0) {
						putlog(1, "HTTP done on upstream fd %d (despite buggy httpd)", str->src.fd);
						str->src.httpdone = 1;
						memmove(&str->pbuf[0], &str->pbuf[j+2], (len - (j + 2)));
						len -= (j + 2);
						break;
					}
				}
			}
			if (!str->src.httpdone) {
				len = 0;
			}
			if (len < 0) {
				len = 0;
			}
		}
		str->pbufvalid += len;
		if ((str->pbufvalid + readsize) >= MAX_MCAST_PAYLOAD) {
			/* Another packet won't fit, so send now */
			send_packet(str, str->pbuf);
			str->pbufvalid = 0;
		}
	} else {
		/* Not sure how to handle this? Probably best to close */
		putlog(1, "Warning: Closed upstream fd %d due to error", fd);
		event_del(&str->src.ev);
		close(fd);
		str->src.isopen = 0;
		str->pbufvalid = 0;
	}
}

void work_tcp_connection(int fd, short event, void * up) {
	struct stream * str = (struct stream *)up;
	char buf[1024];
	
	if (event & EV_READ) {
		int rdres;
		rdres = read(fd, buf, 1023);
		if (rdres <= 0) {
			putlog(1, "[fd %05d]: Read error, dropping connection", fd);
			drop_tcp_connection(str, fd);
			return;
		} else {
			struct unicast_client * uc = str->dst.ucclients;
			while (uc) {
				if ((uc->sockfd == fd) && (!uc->hadrequest)) {
					write(uc->sockfd,
					      HTTP_OK_REPLY,
					      strlen(HTTP_OK_REPLY));
					uc->hadrequest = 1;
				}
				uc = uc->next;
			}
#if 0
			buf[rdres]=0;
			printf("Read: %d bytes: %s\n", rdres, buf);
#endif
		}
	}
}

void new_tcp_connection(int fd, short event, void * up) {
	struct stream *	s = (struct stream *)up;
	struct sockaddr_in remaddr;
	unsigned int sl = sizeof(struct sockaddr_in);
	int newsock;
	struct unicast_client *	uc;
	int sopt;
	
	newsock = accept(fd, (struct sockaddr *)&remaddr, &sl);
	if (newsock < 0) {
		perror("(ignoring error) ACCEPT");
		return;
	}
	openstreamresource(&s->src, 1, s);
	putlog(1, "[fd %05d]: New connection from %s to stream from fd %d",
	          newsock, inet_ntoa(remaddr.sin_addr), s->src.fd);
	if (s->dst.nuccs >= maxuccs) {
		putlog(1, "[fd %05d]: Too many connections already, dropping the new one.",
		          newsock);
		close(newsock);
		return;
	}
	setfdnonblocking(newsock);
	sopt = sndbuf_size;
	if (setsockopt(newsock, SOL_SOCKET, SO_SNDBUF, (char *)&sopt, sizeof(sopt))) {
		perror("setsockopt SNDBUF");
	}
	
	uc = calloc(1, sizeof(struct unicast_client));
	uc->sockfd = newsock;
	uc->next = s->dst.ucclients;
	s->dst.ucclients = uc;
	s->dst.nuccs++;
	
	event_set(&uc->ev, uc->sockfd, EV_READ | EV_PERSIST, work_tcp_connection, s);
	event_add(&uc->ev, NULL);
}

void openstreamresource(struct streamresource * sr, int issrc, struct stream * parent) {
	if (sr->isopen) {
		return;
	}
	if (sr->type == ST_FILE) {
		/* Open the file (which should be a pipe, otherwise it wouldn't make much sense) */
		putlog(1, "Opening stream resource: FILE: %s (issrc = %d)", sr->fname, issrc);
		sr->fd = open(sr->fname, ((issrc) ? O_RDONLY : O_WRONLY) | O_NONBLOCK);
		if (sr->fd < 0) {
			perror("open");
			return;
		}
		if (issrc) {
			event_set(&sr->ev, sr->fd, EV_READ | EV_PERSIST, data_incoming, parent);
			event_add(&sr->ev, NULL);
		}
	} else {
		/* MCast or http - open socket */
		struct sockaddr_in	sin;
		int sopt;
		
		sin.sin_family = AF_INET;
		memcpy((char *)&sin.sin_addr.s_addr, (char *)&sr->ipv4, sizeof(uint32_t));
		sin.sin_port = htons(sr->port);
		
		putlog(1, "Opening stream resource: %s: %s (%s)",
		       (sr->type == ST_HTTP) ? "HTTP" : "MULTICAST",
		       sr->fname,
		       ((issrc) ? "input" : "output"));
		if (sr->type == ST_MCAST) { /* Multicast address */
			sr->fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
			if (sr->fd < 0) {
				perror("socket(UDP)");
				return;
			}
			if (issrc) { /* Data source */
				struct ip_mreq joinreq;
				sopt = 1;
				if (setsockopt(sr->fd, SOL_SOCKET, SO_REUSEADDR, &sopt, sizeof(sopt)) < 0) {
					perror("setsockopt");
				}
				if (bind(sr->fd, (struct sockaddr *) &sin, sizeof(struct sockaddr_in))) {
					perror("bind");
				}
				/* Join multicast group */
				joinreq.imr_interface.s_addr = INADDR_ANY;
				memcpy((char *)&joinreq.imr_multiaddr.s_addr, (char *)&sr->ipv4, sizeof(uint32_t));
				if (setsockopt(sr->fd, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char *)&joinreq, sizeof(joinreq))) {
					perror("setsockopt IP_ADD_MEMBERSHIP");
				}
				setfdnonblocking(sr->fd);
				event_set(&sr->ev, sr->fd, EV_READ | EV_PERSIST, data_incoming, parent);
				event_add(&sr->ev, NULL);
			} else { /* Data destination */
				sopt = sndbuf_size;
				if (setsockopt(sr->fd, SOL_SOCKET, SO_SNDBUF, (char *)&sopt, sizeof(sopt))) {
					perror("setsockopt SNDBUF");
				}
				if (connect(sr->fd, (struct sockaddr *)&sin, sizeof(struct sockaddr_in))) {
					perror("connect");
					return;
				}
				/* Set socket TTL */
				setsockopt(sr->fd, IPPROTO_IP, IP_MULTICAST_TTL, (char *)&mcastttl, sizeof(mcastttl));
			}
		} else { /* Unicast address */
			sr->fd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
			if (sr->fd < 0) {
				perror("socket(TCP)");
				return;
			}
			if (issrc) { /* Data source */
				/* Connect */
				char * filepartofurl;
				if (connect(sr->fd, (struct sockaddr *)&sin, sizeof(struct sockaddr_in))) {
					perror("connect");
					return;
				}
				setfdnonblocking(sr->fd);
				/* Send HTTP request */
				write(sr->fd, HTTP_REQUEST_PART1, strlen(HTTP_REQUEST_PART1));
				filepartofurl = index(sr->fname, '/');
				if (filepartofurl == NULL) {
					filepartofurl = "/";
				}
				write(sr->fd, filepartofurl, strlen(filepartofurl));
				write(sr->fd, HTTP_REQUEST_PART2, strlen(HTTP_REQUEST_PART2));
				sr->httpdone = 0;
				event_set(&sr->ev, sr->fd, EV_READ | EV_PERSIST, data_incoming, parent);
				event_add(&sr->ev, NULL);
			} else { /* Data destination */
				setfdnonblocking(sr->fd);
				sopt = 1;
				if (setsockopt(sr->fd, SOL_SOCKET, SO_REUSEADDR, (char *)&sopt, sizeof(sopt))) {
					perror("setsockopt REUSEADDR");
				}
				if (bind(sr->fd, (struct sockaddr *) &sin, sizeof(struct sockaddr_in))) {
					perror("bind");
					return;
				}
				if (listen(sr->fd, 10)) {
					perror("listen");
					return;
				}
				sr->nuccs = 0;
				event_set(&sr->ev, sr->fd, EV_READ | EV_PERSIST, new_tcp_connection, parent);
				event_add(&sr->ev, NULL);
			}
			
		}
	}
	sr->isopen = 1;
}

void parsestreamparams(char * what, char * whereto) {
	struct stream * tmpstr;

	tmpstr = streams;
	streams = calloc(1, sizeof(struct stream));
	streams->next = tmpstr;
	parseurl(what,    &streams->src);
	parseurl(whereto, &streams->dst);
	if ((readpersist) || (streams->dst.type != ST_HTTP)) {
		openstreamresource(&streams->src, 1, streams);
	}
	openstreamresource(&streams->dst, 0, streams);
}

int main(int argc, char ** argv) {
	char	ch;
	int	argpos;
	
	printf("Streamforwarder Version %s%s Compiled on %s %s\n",
	       "0.1", "+beta0", __DATE__, __TIME__);
	
	init_sigpipehandler();
	event_init();
	
	while ((ch = getopt(argc, argv, "b:d:l:hp")) != -1) {
		switch (ch) {
			case 'b':
				sndbuf_size = strtoul(optarg, NULL, 10);
				break;
			case 'd':
				console_loglevel = strtol(optarg, NULL, 10);
				break;
			case 'h':
				showhelptxt(argv[0]);
				exit(-1);
				break;
			case 'l':
				maxuccs = strtoul(optarg, NULL, 10);
				break;
			case 'p':
				readpersist = 1;
				break;
			default:
				fprintf(stderr, "Unknown command line arg\n");
				fprintf(stderr, "Try %s -h for help\n", argv[0]);
				exit(-1);
				break;
		}
	}
	argpos = optind;
	while (argpos < argc) {
		if ((argc - argpos) < 2) {
			fprintf(stderr, "ERROR: Non-Option arguments always go as pairs: First what to forward, then where to forward it to.\n");
			fprintf(stderr, "In your arguments, '%s' has no place to be forwarded to...\n", argv[argpos]);
			exit(-1);
		}
		parsestreamparams(argv[argpos], argv[argpos+1]);
		argpos += 2;
	}
	
	if (!streams) {
		fprintf(stderr, "ERROR: Need at least one thing to stream\n");
		exit(-1);
	}

	/* Pigs can fly */
	putlog(0, "running.");
	event_dispatch();

	return 0;
}
