
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include "ttyhandlers.h"

#define INPUTBUFSIZE 40
#define BACKLOGSIZE 20
static unsigned char inputbuf[INPUTBUFSIZE * BACKLOGSIZE];
static unsigned int inppos = 0;
static unsigned int blpos = 0;
static unsigned char escstatus = 0;
static tabcompletioncallback * tccb = NULL;
/* Original terminal attributes. */
static struct termios * saved_attributes = NULL;

/* Set up stdin so that we can get input from it
 * Returns 0 if stdin is not a terminal, or 1 if it is.
 * Mostly copied from the GNU LIBC Docs */
int setupstdin(void) {
	struct termios tattr;

	/* Make sure stdin is a terminal. */
	if (!isatty(STDIN_FILENO)) {
		/* Not a terminal, ignore it */
		return 0;
	}

	/* Save the terminal attributes so we can restore them later. */
	saved_attributes = calloc(sizeof(struct termios), 1);
	tcgetattr(STDIN_FILENO, saved_attributes);
	atexit(resetstdin);

	/* Set the funny terminal modes. */
	tcgetattr(STDIN_FILENO, &tattr);
	tattr.c_lflag &= ~(ICANON | ECHO); /* Clear ICANON and ECHO. */
	tattr.c_cc[VMIN] = 1;
	tattr.c_cc[VTIME] = 0;
	tcsetattr(STDIN_FILENO, TCSAFLUSH, &tattr);
	
	return 1;
}

int addtoinputbuf(unsigned char which) {
	int ret = 0;
	if (escstatus == 0) {
		switch(which) {
		case '\t': /* Tab */
			if (tccb == NULL) {
				makebell();
			} else { /* Try */
				unsigned char * r;
				inputbuf[(blpos * INPUTBUFSIZE) + inppos] = 0;
				r = tccb(&inputbuf[blpos * INPUTBUFSIZE]);
				if (r == NULL) {
					makebell();
				} else {
					strncpy(&inputbuf[blpos * INPUTBUFSIZE],
					        r,
					        INPUTBUFSIZE - 1);
					inputbuf[(blpos * INPUTBUFSIZE) + INPUTBUFSIZE - 1] = 0;
					inppos = strlen(&inputbuf[blpos * INPUTBUFSIZE]);
				}
			}
			break;
		case 8:
		case 0x7F: /* Backspace */
			if (inppos > 0) {
				inppos--;
			}
			break;
		case 10:
		case 13:
			if (inppos > 0) {
				inputbuf[(blpos * INPUTBUFSIZE) + inppos] = 0;
				if (blpos > 0) {
					memcpy(&inputbuf[0 * INPUTBUFSIZE],
					       &inputbuf[blpos * INPUTBUFSIZE],
					       INPUTBUFSIZE);
					blpos = 0;
				}
				if (strcmp(&inputbuf[1 * INPUTBUFSIZE],
				           &inputbuf[0 * INPUTBUFSIZE]) != 0) {
					memmove(&inputbuf[2 * INPUTBUFSIZE],
					        &inputbuf[1 * INPUTBUFSIZE],
					        (BACKLOGSIZE - 2) * INPUTBUFSIZE);
					memcpy(&inputbuf[1 * INPUTBUFSIZE],
					       &inputbuf[0 * INPUTBUFSIZE],
					       INPUTBUFSIZE);
				}
				ret = 1;
			}
			break;
		case 27: /* Escape */
			escstatus++;
			break;
		default:
			if (inppos >= (INPUTBUFSIZE - 2)) {
				makebell();
			} else {
				inputbuf[(blpos * INPUTBUFSIZE) + inppos] = which;
				inppos++;
			}
			break;
		};
	} else if (escstatus == 1) { /* Had ESC, expect [ */
		if (which == '[') {
			escstatus++;
		} else {
			escstatus = 0;
		}
	} else if (escstatus == 2) { /* Had ESC[, expect parameter */
		/*printf("| %s: 0x%02x |\n", __FUNCTION__, which);*/
		switch (which) {
		case 0x41: /* Cursor up */
			if ((blpos < (BACKLOGSIZE - 1)) 
			 && (inputbuf[(blpos + 1) * INPUTBUFSIZE] != 0)) {
				blpos++;
				inppos = strlen(&inputbuf[blpos * INPUTBUFSIZE]);
			} else {
				makebell();
			}
			break;
		case 0x42: /* Cursor down */
			if (blpos > 0) {
				blpos--;
				inppos = strlen(&inputbuf[blpos * INPUTBUFSIZE]);
			} else {
				makebell();
			}
			break;
		case 0x43: /* Cursor right */
		case 0x44: /* Cursor left */
		default:
			makebell();
			break;
		};
		escstatus = 0;
	}
	redrawinputline();
	return ret;
}

void redrawinputline(void) {
	if (isatty(STDOUT_FILENO)) {
		jmptocol1();
		cleartoeol();
		inputbuf[(blpos * INPUTBUFSIZE) + inppos] = 0;
		printf("> %s", &inputbuf[blpos * INPUTBUFSIZE]);
		fflush(stdout);
	}
}

void clearinputbuf(void) {
	inppos = 0;
}

unsigned char * getinputline(void) {
	inputbuf[inppos] = 0;
	return &inputbuf[0];
}

void resetstdin(void) {
	if (saved_attributes != NULL) {
		tcsetattr(STDIN_FILENO, TCSANOW, saved_attributes);
		free(saved_attributes);
		saved_attributes = NULL;
	}
}

void makebell(void) {
	if (isatty(STDOUT_FILENO)) {
		printf("%c", 7); /* Bell */
	}
}

void jmptocol1(void) {
	if (isatty(STDOUT_FILENO)) {
		printf("%c[999D", 27);
	}
}

void cleartoeol(void) {
	if (isatty(STDOUT_FILENO)) {
		printf("%c[K", 27);
	}
}

void registertabcompletioncallback(tabcompletioncallback * x) {
	tccb = x;
}
